-------------------------------------------------------------
library IEEE;
use IEEE.STD_LOGIC_1164.ALL;
use IEEE.NUMERIC_STD.ALL;

use work.ipbus.all;
use work.ipbus_reg_types.all;
use work.emp_data_types.all;
use work.mprocessor_constants.all;
use work.ipbus_decode_eprocessor.all;
use work.dtc_link_maps.all;

entity EcalDataProcessor is
port (
       --- Input Ports ---
        clk_p            : in  std_logic;
        clk40            : in  std_logic;
        data_in          : in tDaqEcalLinkDataArray;
        gbe_backpressure : in std_logic;
        --- Output Ports ---
        data_out         : out ldata(cNumberOfEcalOutputLinks - 1 downto 0);
       --- IPBus Ports ---
        ipb_clk            : in  std_logic;
        ipb_rst            : in  std_logic;
        ipb_in             : in  ipb_wbus;
        ipb_out            : out ipb_rbus
);
end EcalDataProcessor;

architecture rtl of EcalDataProcessor is

  signal ipb_to_slaves           : ipb_wbus_array(N_SLAVES - 1 downto 0);
  signal ipb_from_slaves         : ipb_rbus_array(N_SLAVES - 1 downto 0);

  signal ctrl_regs  : ipb_reg_v(0 to 0) := (others => (others => '0'));
  signal stat_regs  : ipb_reg_v(0 to 0);

  signal reset      : std_logic;
  signal latency    : std_logic_vector(8 downto 0);
  
  -- 448 length data with valid(1 clock at 320MHz when the data changes at 40MHz)
  signal data_start                   : std_logic;
  signal data_last                    : std_logic;
  signal data                         : std_logic_vector(28 * 16 -1 downto 0);

  signal trig_lat                     : std_logic_vector(5 downto 0);
  signal trig                         : std_logic_vector(31 downto 0);
  
  signal data_ram_ena      : std_logic;
  signal data_ram_wea      : std_logic;
  signal data_ram_addra    : std_logic_vector(8 DOWNTO 0);
  signal data_ram_dina     : std_logic_vector(66 DOWNTO 0);
  signal data_ram_enb      : std_logic;
  signal data_ram_addrb    : std_logic_vector(8 DOWNTO 0);
  signal data_ram_doutb    : std_logic_vector(66 DOWNTO 0);

  signal trig_ram_ena      : std_logic;
  signal trig_ram_wea      : std_logic;
  signal trig_ram_addra    : std_logic_vector(5 downto 0);
  signal trig_ram_dina     : std_logic_vector(31 DOWNTO 0);
  signal trig_ram_enb      : std_logic;
  signal trig_ram_addrb    : std_logic_vector(5 downto 0);
  signal trig_ram_doutb    : std_logic_vector(31 DOWNTO 0);
  
	signal trig_ram_doutb_l  : std_logic_vector(31 DOWNTO 0);
  
  signal trig_ram_enb_p      : std_logic_vector(1 downto 0);
  signal trig_valid_p      : std_logic_vector(2 downto 0);
  signal data_out_p        : ldata(2 downto 0);
  
  signal data_ready_i      : std_logic;
  signal d_waddr_valid_i   : std_logic;
  signal t_waddr_valid_i   : std_logic;
  signal raddr_valid_i     : std_logic; 
  
  signal trig_lat_offset   : std_logic_vector(8 downto 0);

  signal data_out_s        :  lword;

  type ostate_type is (wait_start, wait_last, wait_bpoff); -- This state machine does not allow a single frame packet. The packet length >= 2.
  signal ostate            : ostate_type;
   
begin

  --==============================--
  fabric: entity work.ipbus_fabric_sel
  --==============================--
  generic map(
    NSLV            => N_SLAVES,
    SEL_WIDTH       => IPBUS_SEL_WIDTH
  )
  port map(
        ipb_in          => ipb_in,
        ipb_out         => ipb_out,
        sel             => ipbus_sel_eprocessor(ipb_in.ipb_addr),
        ipb_to_slaves   => ipb_to_slaves,
        ipb_from_slaves => ipb_from_slaves
  );

  --==============================--
  ctrl_reg: entity work.ipbus_ctrlreg_v
  --==============================--
  generic map(
        N_CTRL    => 1,
        N_STAT    => 1
  )
  port map(
        clk       => ipb_clk,
        reset     => ipb_rst,
        ipbus_in  => ipb_to_slaves(N_SLV_DP),
        ipbus_out => ipb_from_slaves(N_SLV_DP),
        d         => stat_regs,
        q         => ctrl_regs
  );
  reset <= ctrl_regs(0)(0);
  latency <= ctrl_regs(0)(9 downto 1);
  trig_lat_offset <= ctrl_regs(0)(18 downto 10);
  stat_regs(0)(8 downto 0) <= latency;
  stat_regs(0)(17 downto 9) <= trig_lat_offset;
  stat_regs(0)(18) <= gbe_backpressure;
  --  stat_regs(0)(10) <= d_waddr_valid_i;
  -- stat_regs(0)(11) <= t_waddr_valid_i;
  -- stat_regs(0)(12) <= raddr_valid_i;
  -- stat_regs(0)(13) <= zsf_bypass;
 -- stat_regs(0)(17) <= trig_ram_dina(0);
 -- stat_regs(0)(18) <= trig_ram_doutb(0);
 -- stat_regs(0)(19) <= trig_valid_p(0);
 -- stat_regs(0)(20) <= trig_valid_p(1); 
 -- stat_regs(0)(20) <= trig_valid_p(2); 
       
  --==============================--
  EcalDataAggregatorInstance : entity work.EcalDataAggregator
  --==============================--
  port map (
    clk_p          => clk_p,
    data_in        => data_in,
    data_out_start => data_start,
    data_out_last  => data_last,
    data_out       => data
  );

  --==============================--
  EcalDataFilterInstance : entity work.EcalDataFilter
  --==============================--
  port map (
    clk_p              => clk_p,
    clk40              => clk40,
    reset              => reset,
    data_in_start      => data_start,
    data_in            => data,
    trig_o_lat         => trig_lat,
    trig_o             => trig,
    --- IPBus Ports ---
    ipb_clk       => ipb_clk,
    ipb_rst       => ipb_rst,
    ipb_in        => ipb_to_slaves(N_SLV_DFilter),
    ipb_out       => ipb_from_slaves(N_SLV_DFilter)
  );


--  --==============================--
--  EcalDataMonitorInstance : entity work.EcalDataMonitor
--  --==============================--
--  port map (
--    clk_p              => clk_p,
--    data_in_valid      => data_valid,
--    data_in_all_valid  => data_all_valid,
--    data_in            => df_data,
--    --- IPBus Ports ---
--    ipb_clk       => ipb_clk,
--    ipb_rst       => ipb_rst,
--    ipb_in        => ipb_to_slaves(N_SLV_MON),
--    ipb_out       => ipb_from_slaves(N_SLV_MON)
--  );
  
  --==============================--
  ecal_data_ram : entity work.ecal_data_ram
  --==============================--
  PORT MAP (
    clka  => clk_p,
    ena   => data_ram_ena,
    wea(0)=> data_ram_wea,
    addra => data_ram_addra,
    dina  => data_ram_dina,
    clkb  => clk_p,
    enb   => data_ram_enb,
    addrb => data_ram_addrb,
    doutb => data_ram_doutb
  );
  --==============================--
  ecal_trig_ram : entity work.ecal_trig_ram
  --==============================--
  PORT MAP (
    clka  => clk_p,
    ena   => trig_ram_ena,
    wea(0)=> trig_ram_wea,
    addra => trig_ram_addra,
    dina  => trig_ram_dina,
    clkb  => clk_p,
    enb   => trig_ram_enb,
    addrb => trig_ram_addrb,
    doutb => trig_ram_doutb
  ); 
  
  bp: process(clk_p)
		constant FMAX  : integer := 255;
		variable fsize : integer := 0;
  begin
        if reset = '1' then
					fsize := 0;
          data_out(0).data  <= (others => '0');
          data_out(0).valid <= '0';
          data_out(0).start <= '0';
          data_out(0).last <= '0';
          data_out(0).strobe <= '0';
          ostate <= wait_start;
        elsif rising_edge(clk_p) then

          data_out(0).data  <= data_out_s.data;
          data_out(0).valid <= '0';
          data_out(0).start <= '0';
          data_out(0).last <= '0';
          data_out(0).strobe <= '0';
          case ostate is
            when wait_start =>
							fsize := 0;
              if data_out_s.start = '1' and data_out_s.valid = '1' then
								fsize := 1;
                data_out(0).start <= data_out_s.start;
                data_out(0).valid <= data_out_s.valid;
                ostate <= wait_last;
							else
								ostate <= wait_start;
              end if;
            when wait_last =>
                data_out(0).valid <= data_out_s.valid;
								fsize := fsize + 1;
								if data_out_s.valid = '0' or fsize = FMAX then
                  data_out(0).valid <= '1';
                  data_out(0).last <= '1';
                  data_out(0).strobe <= '1';
          				data_out(0).data  <= (others => '1');
                  if gbe_backpressure = '0' then 
                    ostate <= wait_start; 
                  else
                    ostate <= wait_bpoff;
                  end if;
                elsif data_out_s.last = '1' then
                  data_out(0).valid <= data_out_s.valid;
                  data_out(0).last <= data_out_s.last;
                  data_out(0).strobe <= data_out_s.strobe;
                  if gbe_backpressure = '0' then 
                    ostate <= wait_start; 
                  else
                    ostate <= wait_bpoff;
                  end if;
                end if;
            when wait_bpoff =>
              if gbe_backpressure = '0' then
                ostate <= wait_start;
              end if;
          end case;

        end if;
  end process;

  ram_rw: process(clk_p)
    variable data_ready : boolean := false;
    variable f_timer : integer range 0 to 7 := 0;
    variable g_timer : integer range 0 to 7 := 0;
    -- for writing data ram
    variable d_waddr : integer range 0 to 511 := 0;
    variable d_waddr_valid : boolean := false;
    variable d_waddr_wcount : integer range 0 to 8 := 0;
    -- for writing trig ram
    variable t_waddr : integer range 0 to 63 := 0; 
    variable t_waddr_valid : boolean := false;
    variable t_waddr_wcount : integer range 0 to 63 := 0;
    -- for reading rams
    variable raddr : integer range 0 to 511 := 0;
    variable raddr_valid : boolean := false;
    variable raddr_wcount : integer range 0 to 511 := 0;
    
    -- data ram
    constant bit_valid : integer := 64;
    constant bit_start : integer := 65;
    constant bit_last  : integer := 66;
    begin
  
        if reset = '1' then
          data_ready := false;
          d_waddr := 0;
          t_waddr := 0;
          raddr := 0;      
          data_ram_ena   <= '0';
          data_ram_wea   <= '0';
          data_ram_addra <= (others => '0');
          data_ram_dina  <= (others => '0');
          data_ram_enb   <= '0';
          data_ram_addrb <= (others => '0');
          trig_ram_ena   <= '0';
          trig_ram_wea   <= '0';          
          trig_ram_addra <= (others => '0');
          trig_ram_dina  <= (others => '0');
          trig_ram_enb   <= '0';
          trig_ram_enb_p <= (others => '0');
          trig_ram_addrb  <= (others => '0');

          g_timer        := 0;
          d_waddr_valid  := false;
          d_waddr_wcount := 0;
          t_waddr_valid  := false;
          t_waddr_wcount := 0;
          raddr_valid    := false;
          raddr_wcount   := 0;
        elsif rising_edge(clk_p) then
        
          trig_ram_enb_p(0) <= trig_ram_enb;
          trig_ram_enb_p(1) <= trig_ram_enb_p(0);
          if data_ready then
            data_ready_i <= '1';
          else 
            data_ready_i <= '0';
          end if;
          if d_waddr_valid then
            d_waddr_valid_i <= '1'; 
          else 
            d_waddr_valid_i <= '0';
          end if;
          if t_waddr_valid then
            t_waddr_valid_i <= '1';
          else
            t_waddr_valid_i <= '0';
          end if;
          if raddr_valid then
            raddr_valid_i   <= '1';
          else 
            raddr_valid_i   <= '0';
          end if;
          
          if d_waddr_valid then
            data_ram_ena   <= '1';
            data_ram_wea   <= '1';
            data_ram_addra <= std_logic_vector(to_unsigned(d_waddr,9));
            trig_ram_ena    <= '1';           
          end if;
 
          trig_ram_wea    <= '0';   
          if g_timer = 0 then           
            if t_waddr_valid then
              trig_ram_wea   <= '1';
              trig_ram_addra <= std_logic_vector(to_unsigned(t_waddr,6));
            end if;
          end if;  
 
          data_out_s.data  <= (others => '0');
          data_out_s.valid <= '0';
          data_out_s.start <= '0';
          data_out_s.strobe <= '0';
          if trig_ram_enb = '1' and trig_ram_enb_p(1) = '1' then
 
            data_ram_addrb <= std_logic_vector(to_unsigned(raddr,9));
            trig_ram_addrb <= std_logic_vector(to_unsigned(raddr,9)(8 downto 3));
            
            data_out_s.data  <= data_ram_doutb(63 downto 0);
            if data_ram_doutb(bit_start) = '1' then
              data_out_s.data(63 downto 32) <= trig_ram_doutb;
							trig_ram_doutb_l              <= trig_ram_doutb;
              data_out_s.start <= trig_ram_doutb(31);
              data_out_s.valid <= trig_ram_doutb(0) and data_ram_doutb(bit_valid);           
						else
              data_out_s.valid <= trig_ram_doutb_l(0) and data_ram_doutb(bit_valid);           
              data_out_s.last  <= trig_ram_doutb_l(30) and data_ram_doutb(bit_last);
              data_out_s.strobe <= trig_ram_doutb_l(30) and data_ram_doutb(bit_last);         
            end if;
          end if;
          
          -- data to write                   
          data_ram_dina(63 downto 0) <= (others => '0');
          data_ram_dina(bit_valid) <= '0';
          data_ram_dina(bit_start) <= '0';
          data_ram_dina(bit_last)  <= '0';
          if data_start = '1' then
            if data_ready = false then
              g_timer := 0;
              data_ready := true;
            end if;
            -- data frame 0 , timer 7
            data_ram_dina(31 downto 0) <= data(data'high - 16 downto data'high - 16 - 31); -- oc? super id?
            data_ram_dina(bit_valid) <= '1';
            data_ram_dina(bit_start) <= '1';
            -- data frame 0
           f_timer := 0;
          else
            if f_timer = 7 then
              f_timer := 7;
            else
              -- data frame 1 - 7, timer 0 - 6
              if f_timer = 0 then
                data_ram_dina(63 downto 32) <= data(data'high - 16 downto data'high - 16 - 31); -- calib
                data_ram_dina(31 downto 16) <= data(data'high - 48 downto data'high - 48 - 15); -- channel 24
                data_ram_dina(12) <= data(data'high - 3); -- spill
                data_ram_dina(11 downto 0) <=  data(data'high - 4 downto data'high - 4 - 11); -- bcid
              elsif f_timer < 7 then
                data_ram_dina(63 downto 0) <= data(64*(7-f_timer)-1 downto 64*(6-f_timer));
              end if;
              data_ram_dina(bit_valid) <= '1';
              f_timer := f_timer + 1;
              if f_timer = 7 then
                data_ram_dina(bit_last) <= '1';
              end if;
              -- data frame 1 - 7
            end if; -- f_timer 
          end if; -- data_valid

          -- trigger to write 
          trig_ram_dina  <= trig;

          -- data ram write address counters
          if data_ready then
            if d_waddr_valid = false then
              d_waddr_wcount := d_waddr_wcount + 1;
            else
              if d_waddr = 511 then
                d_waddr := 0;
              else
                d_waddr := d_waddr + 1;
              end if;
            end if;
            if d_waddr_wcount = 8 then -- to get the data events in the address [8 * i:8 * i + 7]
              d_waddr_valid := true;
            end if; 
          end if;

          -- read rams
          if t_waddr_valid then
            if raddr_valid = false then
              raddr_wcount := raddr_wcount + 1;
            else
              if raddr = 511 then
                raddr := 0;
              else
                raddr := raddr + 1;
              end if;
            end if;
          
            if raddr_wcount = to_integer(unsigned(latency)) then
              raddr_valid := true;
              data_ram_enb <= '1';
              trig_ram_enb <= '1';
            end if;
          end if;
            
          -- trigger ram write data            
          if g_timer = 0 then
            
            if d_waddr_valid then
              
              if t_waddr_valid = false then
                t_waddr_wcount := t_waddr_wcount + 1;
              else        
                if t_waddr = 63 then
                  t_waddr := 0;
                else
                  t_waddr := t_waddr + 1;
                end if;
              end if;
            
              if t_waddr_wcount = to_integer(unsigned(trig_lat)) + to_integer(unsigned(trig_lat_offset)) then
                t_waddr_valid := true;
              end if; 
            end if;
          end if;
          
          if g_timer = 7 then
            g_timer := 0;        
          else
            g_timer := g_timer + 1;
          end if;
            
        end if;   
    end process;
    
end rtl;
