----------------------------------------------------------------------------------
-- ZeroSuppression filter module.
-- Designed for MuOne experiment - August 2023 
-- Fabio Montecassiano - INFN Padova

--- TODO
--- a lot...
--

-- MEMO ------------------------------------------
--  a: short name for 'no. of samples AFTER trg has been asserted'
--  b: short name for 'no. of samples BEFORE trg will be asserted'
--  c: short name for 'no. of samples to be CHECKED vs. threshold in order to assert the trg signal'


library ieee;
use ieee.std_logic_1164.all;
use ieee.numeric_std.all;

use ieee.math_real.all; -- for static elaboration of log2()
use WORK.ecal_zsf_type_pckg.all;

entity EcalDataZSFilter is
generic(
  N_ch:   POSITIVE:=  27;      -- Sets the no. of channels (Ecal August 2023 requires  27ch=25+2 calib ch)
  DeadTime_MAX:
           POSITIVE:= 2**16-1; -- Set the the limit for the DeadTime between 2 consecutive 'valid' assertion in in terms of no. of clocks
  --
  -- NB The generics "a_LEFT, b_LEFT, c_LEFT" control the sharing of the slow control register 'sc_regs_i(28)' among the 3 parameters Nsmpl_a, Nsmpl_b, Nsmpl_c
  --
  a_MAX:  POSITIVE:= 2**8-1;  -- Sets the limit of the observable future in in terms of no. of samples after the C1 assertion when in 'BASIC mode'
  a_LEFT: NATURAL:= 16;       -- Sets the lower bound of the index inside slow ctrl reg 'sc_regs_i(28)' for parameter Nsmpl_a
 
  b_MAX:  POSITIVE:= 2**5-1;      -- A.K.A. 'D' deepth of delay line (they must match!). Sets the limit of the observable past in terms of no. of samples before C1 assertion
  b_LEFT: NATURAL:=  8;       -- Sets the lower bound of the index inside slow ctrl reg 'sc_regs_i(28)' for parameter Nsmpl_b
 
  c1_MAX: POSITIVE:= 2**4-1;  -- Sets the limit for the no. of consecutive checks "over threshold"
  c_LEFT: NATURAL:=  0;       -- Sets the lower bound of the index inside slow ctrl reg 'sc_regs_i(28)' for parameter Nsmpl_c

  ch_trg_MAX:   POSITIVE:= 32  -- Maximum no. of clocks after that the deassertion of the channel trigger signal 'ch_trgs()' is forced
);
port
(
  clk:              in std_logic;  -- 40 MHz
  rst:              in std_logic;
  
  ----------------------------------------------------------------
  -- PORT 'lpgbt_payload_i' assumes the following map on 56 byte.
  -- IT NEEDS TO BE REMAPPED ACCORDING TO THE SERENITY FW                  <<<<<<<<<<<<<<<<<<<< TODO
  --  lpgbt_payload_i(447..445):     reserved/sync     [ 3 bit]
  --  lpgbt_payload_i(444):          spill_mskd        [ 1 bit]
  --  lpgbt_payload_i(443..432):     bc                [12 bit]
  --  lpgbt_payload_i(431..400):     oc_OR_calib_ch    [32 bit]
  --  lpgbt_payload_i(399..0):       adc_data_i(24)..adc_data_i(0) [25x16 bit]
  lpgbt_payload_i:  in std_logic_vector(447 downto 0); 
  ----------------------------------------------------------------
  -- Slow control operated through an array of ipbus registers delivering
  --   threshold 27x16bit,  mask_trg_ch 1x 27bit, ...
  -- Provisional map of each of 32 bits register:
  --     0: thrsld_ch(0)  [highest 16 b RESERVED / lowest 16 b 'threshold physic ch #0']
  --    ...
  --    24: thrsld_ch(24) [highest 16 b RESERVED / lowest 16 b 'threshold physic ch #24']
  --    25: thrsld_clbr(0)[highest 16 b RESERVED / lowest 16 b 'threshold calibration ch #0']
  --    26: thrsld_clbr(1)[highest 16 b RESERVED / lowest 16 b 'threshold calibration ch #1']
  --    27: mask_trg_ch   [highest  5 b RESERVED / lowest 27 b 'mask_trg_ch']
  --    28: Nsmpl_*       [highest  8 b RESERVED / 8 b 'Nsmpl_a' / 8 b 'Nsmpl_b' / 4 b RESERVED / 4 b 'Nsmpl_c']
  --    29: DeadTime      [highest 16 b RESERVED / lowest 16 b 'DeadTime']
  --    30: RESERVED
  --    31: Control       [highest 29 b RESERVED / 2:'en_adv'/ 1:'en_dtime' / 0:'en_rtrg']
  sc_regs_i:        in  array_32bit(0 to 31); -- array (0 to 31) of std_logic_vector(31 downto 0) --std_logic_vector(1023 downto 0);
  valid_o:          out std_logic;
  all_valids_o:     out std_logic_vector(N_ch-1 downto 0);
  valid_lat_o:      out std_logic_vector(5 downto 0)
); 
end entity;
 
architecture rtl of EcalDataZSFilter is

  -- Derived constants ----------------------------------
  constant Nsmpl_c_BITS: POSITIVE:=positive(ceil(log2(real(c1_MAX))));
  constant Nsmpl_a_BITS: POSITIVE:=positive(ceil(log2(real(a_MAX))));
  constant vld_MAX: POSITIVE:= a_MAX+b_MAX+1;
  constant Nsmpl_b_BITS: POSITIVE:=positive(ceil(log2(real(b_MAX))));
  constant DeadTime_BITS:POSITIVE:=positive(ceil(log2(real(DeadTime_MAX))));
  
  ------- Slow control by IpBus registers
  signal Nsmpl_c: unsigned(Nsmpl_c_BITS-1 downto 0);
  signal Nsmpl_b: unsigned(Nsmpl_b_BITS-1 downto 0);
  signal Nsmpl_a: unsigned(Nsmpl_a_BITS-1 downto 0);
  signal mask_trg_ch: std_logic_vector(N_ch-1 downto 0);
  signal DeadTime: unsigned(DeadTime_BITS-1 downto 0);
  signal en_dtime, en_rtrg, en_adv: std_logic;
  --
  -- State Machine checking threshold, one per each channel --
  -- States
  type t_State_ch is (WAIT_T, COUNT_C1, ASSRT_CH_TRG);
  -- Counters
  type r_cnt_ch is
  record
       c1:   NATURAL range 0 to c1_MAX+ch_trg_MAX+1;
       trg:  NATURAL range 0 to ch_trg_MAX+1;
  end record;
  -- Channel signals we want to observe as group/array
  type t_array_cnt_ch   is array (integer range <>) of r_cnt_ch;
  type t_array_16bit    is array (integer range <>) of std_logic_vector(15 downto 0);
  type t_array_State_ch is array (integer range <>) of t_State_ch;
  signal ch:        t_array_16bit(N_ch-1 downto 0);
  signal T:         std_logic_vector(N_ch-1 downto 0):=(others=>'0');
  signal thrsld_ch: t_array_16bit(N_ch-1 downto 0);
  signal State_ch, NxtState_ch : t_array_State_ch(N_ch-1 downto 0);
  signal Cnt_ch, NxtCnt_ch:      t_array_cnt_ch(N_ch-1 downto 0);
  -- Output of channels' State Machines
  signal trgs_ch: std_logic_vector(N_ch-1 downto 0);  
  -- 
  procedure BuildNxtState_ch(
    constant State:        in t_State_ch;
    signal   NxtState_ch: out t_State_ch;
    signal   Cnt_ch:       in r_cnt_ch;
    signal   NxtCnt_ch:   out r_cnt_ch;
    variable ch_trg:  out std_logic
  ) is
  begin
    NxtState_ch <= State;
    NxtCnt_ch   <= Cnt_ch;
    case State is
        when WAIT_T =>
            ch_trg:='0';
            NxtCnt_ch.c1 <=0;
            NxtCnt_ch.trg<=0;
        when COUNT_C1 =>
            ch_trg:='0';
            NxtCnt_ch.c1 <=Cnt_ch.c1+1;
            NxtCnt_ch.trg<=0;
        when ASSRT_CH_TRG =>
            ch_trg:='1';
            NxtCnt_ch.c1 <=Cnt_ch.c1+1;
            NxtCnt_ch.trg<=Cnt_ch.trg+1;  
        when OTHERS =>
           REPORT "FAULT CONDITION in 'BuildNextState_ch(NxtState_ch<=<'" & t_State_ch'IMAGE(State) &">,....)"
                  SEVERITY FAILURE;
      end case;    
  end procedure BuildNxtState_ch;
  --
  -- Common State Machine elaborating the valid signal, one for all channels
  -- States
  type t_State_com is (WAIT_TRG, CNT_BFR, TRG_ASSRTD, CNT_AFTR_TRG, WAIT_DTIME);
  signal State_com, NextState_com: t_State_com;
  -- Input of common State Machine
  signal trg: std_logic:='0';
  -- Counters
  type r_cnt_com is
    record
      bfrvld:   NATURAL range 0 to b_MAX+1+1;
      aftrtrg:  NATURAL range 0 to a_MAX+1;
      dtime:    NATURAL range 0 to 2**DeadTime'LENGTH; --TODO check this vs. +1
      vld:      NATURAL range 0 to vld_MAX+1;
    end record;
  signal Cnt_com, NxtCnt_com:      r_cnt_com;

------------------------------------------------------
begin
  valid_lat_o <= std_logic_vector(to_unsigned(b_MAX, 6));
  --
  -- TODO: add assert checks for size of sc register VS. IpBus reserved widths
  --
 
  -- Decode slow control IpBus registers
  mask_trg_ch <= sc_regs_i(27)(mask_trg_ch'RANGE);
  Nsmpl_c <= unsigned(sc_regs_i(28)(c_LEFT+Nsmpl_c_BITS-1 downto c_LEFT));
  Nsmpl_b <= unsigned(sc_regs_i(28)(b_LEFT+Nsmpl_b_BITS-1 downto b_LEFT));
  Nsmpl_a <= unsigned(sc_regs_i(28)(a_LEFT+Nsmpl_a_BITS-1 downto a_LEFT));

  DeadTime<= unsigned(sc_regs_i(29)(DeadTime'RANGE));
  en_rtrg <= sc_regs_i(31)(0);
  en_dtime<= sc_regs_i(31)(1);
  en_adv  <= sc_regs_i(31)(2);

  ----------------------------------------------------
  -- Channels State Machines
  ----------------------------------------------------
SMs_ch: 
  FOR I in 0 to N_ch-1 GENERATE
  BEGIN
    -- Check ADC threshold
    ch(I)  <= lpgbt_payload_i((I+1)*16-1 downto I*16);
    thrsld_ch(I) <= sc_regs_i(I)(15 downto 0);
    T(I)<= '1' when ch(I) > thrsld_ch(I) else '0';
    -- State registers --
  p_regs:process(Clk) is
    begin
      if rising_edge(clk) then
        if rst = '1' then
            State_ch(I) <= WAIT_T;
            Cnt_ch(I)   <= (others => 0);
        else
            State_ch(I) <= NxtState_ch(I);
            Cnt_ch(I)   <= NxtCnt_ch(I);
        end if;
      end if;
    end process;
    
  p_comb:process(T(I), State_ch(I), Cnt_ch(I)) is
    variable ch_trg: std_logic;
    -- Local PROCEDURE to facilitate signal updating   
    procedure BuildNxtState_ch(
      constant StateToSet:   in t_State_ch;
      signal   NxtState_ch: out t_State_ch;
      signal   Cnt_ch:       in r_cnt_ch;
      signal   NxtCnt_ch:   out r_cnt_ch;
      variable ch_trg:      out std_logic
    ) is
    begin
      NxtState_ch <= StateToSet;
      NxtCnt_ch   <= Cnt_ch;
      case StateToSet is
          when WAIT_T =>
              ch_trg:='0';
              NxtCnt_ch.c1 <=0;
              NxtCnt_ch.trg<=0;
          when COUNT_C1 =>
              ch_trg:='0';
              NxtCnt_ch.c1 <=Cnt_ch.c1+1;
              NxtCnt_ch.trg<=0;
          when ASSRT_CH_TRG =>
              ch_trg:='1';
              NxtCnt_ch.c1 <=Cnt_ch.c1+1;
              NxtCnt_ch.trg<=Cnt_ch.trg+1;  
          when OTHERS =>
             REPORT "FAULT CONDITION in 'BuildNxtState_ch(NxtState_ch<=<'" & t_State_ch'IMAGE(StateToSet) &">,....)"
                    SEVERITY FAILURE;
        end case;    
    end procedure BuildNxtState_ch;

    begin -- p_comb
      NxtState_ch(I) <= State_ch(I);
      NxtCnt_ch(I)   <= Cnt_ch(I);
      case State_ch(I) is
        when WAIT_T =>
            if T(I)='0' then
                BuildNxtState_ch(WAIT_T, NxtState_ch(I), Cnt_ch(I), NxtCnt_ch(I), ch_trg);
            elsif (T(I)='1' and Nsmpl_c/=1) then
                BuildNxtState_ch(COUNT_C1, NxtState_ch(I), Cnt_ch(I), NxtCnt_ch(I), ch_trg);
            elsif (T(I)='1' and Nsmpl_c=1) then
                BuildNxtState_ch(ASSRT_CH_TRG, NxtState_ch(I), Cnt_ch(I), NxtCnt_ch(I), ch_trg);
            else
                REPORT LF & "FAULT CONDITION in process 'SMs_ch.p_comb' with State_ch='" & t_State_ch'IMAGE(State_ch(I)) &",....)" & LF SEVERITY FAILURE;
                BuildNxtState_ch(WAIT_T, NxtState_ch(I), Cnt_ch(I), NxtCnt_ch(I), ch_trg);
            end if;                       
        when COUNT_C1 =>
            if T(I)='0' then
                BuildNxtState_ch(WAIT_T, NxtState_ch(I), Cnt_ch(I), NxtCnt_ch(I), ch_trg);
            elsif (T(I)='1' and Cnt_ch(I).c1 < Nsmpl_c) then
                BuildNxtState_ch(COUNT_C1, NxtState_ch(I), Cnt_ch(I), NxtCnt_ch(I), ch_trg);
            elsif (T(I)='1' and Cnt_ch(I).c1 >= Nsmpl_c) then
                BuildNxtState_ch(ASSRT_CH_TRG, NxtState_ch(I), Cnt_ch(I), NxtCnt_ch(I), ch_trg);              
            else
                REPORT LF & "FAULT CONDITION in process 'SMs_ch.p_comb' with State_ch='" & t_State_ch'IMAGE(State_ch(I)) &",....)" & LF SEVERITY FAILURE;
                BuildNxtState_ch(WAIT_T, NxtState_ch(I), Cnt_ch(I), NxtCnt_ch(I), ch_trg);
            end if;
        when ASSRT_CH_TRG => 
            if (T(I)='0' or (T(I)='1' and Cnt_ch(I).trg >= ch_trg_MAX)) then
                BuildNxtState_ch(WAIT_T, NxtState_ch(I), Cnt_ch(I), NxtCnt_ch(I), ch_trg);
            elsif (T(I)='1' and Cnt_ch(I).trg < ch_trg_MAX) then
                BuildNxtState_ch(ASSRT_CH_TRG, NxtState_ch(I), Cnt_ch(I), NxtCnt_ch(I), ch_trg);                            
            else
                REPORT LF & "FAULT CONDITION in process 'SMs_ch.p_comb' with State_ch='" & t_State_ch'IMAGE(State_ch(I)) &",....)" & LF SEVERITY FAILURE;
                BuildNxtState_ch(WAIT_T, NxtState_ch(I), Cnt_ch(I), NxtCnt_ch(I), ch_trg);
            end if;
        when OTHERS =>
            REPORT LF & "FAULT CONDITION in process 'SMs_ch.p_comb' with State_ch='" & t_State_ch'IMAGE(State_ch(I)) &",....)" & LF SEVERITY FAILURE;          
            BuildNxtState_ch(WAIT_T, NxtState_ch(I), Cnt_ch(I), NxtCnt_ch(I), ch_trg);
      end case;    
      -- output of the SM
      trgs_ch(I) <= ch_trg; 
    end process;
    ------------------------------------------
  END GENERATE;

  ----------------------------------------------------
  -- Common State Machine
  ----------------------------------------------------
  -- Input of common SM
  trg <= '0' when  (unsigned(mask_trg_ch(trgs_ch'RANGE) AND trgs_ch) = 0) else '1';   

p_SM_com_reg:  process(clk) is
  begin
      if rising_edge(clk) then
        if rst = '1' then
            State_com <= WAIT_TRG ;
            Cnt_com   <= (others => 0);  
        else
            State_com <= NextState_com;
            Cnt_com   <= NxtCnt_com;
        end if;
      end if;
  end process;
  -----
p_SM_com_comb:process(State_com, trg, Cnt_com) is
  variable vld:   std_logic;
  variable CONDITION:  boolean;
  -- Local PROCEDURE to facilitate signal updating
  -- NB being local to the process, its arguments can be simplified 
  procedure BuildNextState_com(
      constant StateToSet:   in t_State_com;
      signal   NxtState_com:out t_State_com;
      signal   Cnt_com:      in r_cnt_com;
      signal   NxtCnt_com:  out r_cnt_com;
      variable vld:         out std_logic
  ) is
  begin
    NxtState_com <= StateToSet;
    NxtCnt_com   <= Cnt_com;
    case StateToSet is
      when WAIT_TRG =>
        vld:='0';
        NxtCnt_com.vld    <= 0;
        NxtCnt_com.bfrvld <= 0;
        NxtCnt_com.aftrtrg<= 0;
        NxtCnt_com.dtime  <= 0;                       
      when CNT_BFR =>
        vld:='0';
        NxtCnt_com.vld    <= 0;
        NxtCnt_com.bfrvld <= Cnt_com.bfrvld+1;
        NxtCnt_com.aftrtrg<= 0;
        NxtCnt_com.dtime  <= 0;
      when TRG_ASSRTD =>
        vld     :='1';
        NxtCnt_com.vld     <= Cnt_com.vld+1;
        NxtCnt_com.aftrtrg <= 0;
        NxtCnt_com.dtime   <= 0;
      when CNT_AFTR_TRG =>
        vld      :='1';
        NxtCnt_com.vld     <= Cnt_com.vld+1;
        NxtCnt_com.aftrtrg <= Cnt_com.aftrtrg+1;
        NxtCnt_com.dtime   <= 0;
      when WAIT_DTIME =>
        vld:='0';
        NxtCnt_com.dtime   <= Cnt_com.dtime+1;           
      when OTHERS =>
         REPORT "FAULT CONDITION in 'BuildNextState_com(NxtState_com<=<'" & t_State_com'IMAGE(StateToSet) &">,....)"
         SEVERITY FAILURE;
      end case;    
    end procedure BuildNextState_com;    
      
    begin -- p_SM_com_comb
      NextState_com<= State_com;
      NxtCnt_com   <= Cnt_com;
      case State_com is
        when WAIT_TRG =>
          if trg='0' then
            BuildNextState_com(WAIT_TRG, NextState_com, Cnt_com, NxtCnt_com, vld);              
          elsif trg='1' then
            BuildNextState_com(CNT_BFR, NextState_com, Cnt_com, NxtCnt_com, vld);                 
          else
            REPORT LF & "FAULT CONDITION in process 'p_SM_com_comb' with State_com='" & t_State_com'IMAGE(State_com) &",....)" & LF SEVERITY FAILURE;
            BuildNextState_com(WAIT_TRG, NextState_com, Cnt_com, NxtCnt_com, vld);                 
          end if;                       
        when CNT_BFR =>
          CONDITION := Cnt_com.bfrvld<b_MAX-Nsmpl_b;
          if CONDITION then
            BuildNextState_com(CNT_BFR, NextState_com, Cnt_com, NxtCnt_com, vld);                 
          elsif not(CONDITION) then
            BuildNextState_com(CNT_AFTR_TRG, NextState_com, Cnt_com, NxtCnt_com, vld);                 
--          elsif (not(CONDITION) and not(trg='0' or en_adv='0')) then
--            BuildNextState_com(TRG_ASSRTD, NextState_com, Cnt_com, NxtCnt_com, vld);                 
--          elsif (not(CONDITION) and    (trg='0' or en_adv='0')) then
--            BuildNextState_com(CNT_AFTR_TRG, NextState_com, Cnt_com, NxtCnt_com, vld);                 
          else
            REPORT LF & "FAULT CONDITION in process 'p_SM_com_comb' with State_com='" & t_State_com'IMAGE(State_com) &",....)" & LF SEVERITY FAILURE;
            BuildNextState_com(WAIT_TRG, NextState_com, Cnt_com, NxtCnt_com, vld);                 
          end if;  
        when TRG_ASSRTD =>
          CONDITION := Cnt_com.vld<vld_MAX;
          if (trg='1' and CONDITION) then
            BuildNextState_com(TRG_ASSRTD, NextState_com, Cnt_com, NxtCnt_com, vld);                 
          elsif trg='0' then
            BuildNextState_com(CNT_AFTR_TRG, NextState_com, Cnt_com, NxtCnt_com, vld);                 
          elsif (en_dtime='1' and not(CONDITION)) then
            BuildNextState_com(WAIT_DTIME, NextState_com, Cnt_com, NxtCnt_com, vld);                 
          elsif (en_dtime='0' and not(CONDITION)) then
            BuildNextState_com(WAIT_TRG, NextState_com, Cnt_com, NxtCnt_com, vld);                 
          else
            REPORT LF & LF & "FAULT CONDITION in process 'p_SM_com_comb' with State_com='" & t_State_com'IMAGE(State_com) &",....)" & LF SEVERITY FAILURE; 
            BuildNextState_com(WAIT_TRG, NextState_com, Cnt_com, NxtCnt_com, vld);                 
          end if;
        when CNT_AFTR_TRG =>
          CONDITION := ((en_adv='0' and Cnt_com.aftrtrg<Nsmpl_b+Nsmpl_a) or (en_adv='1' and Cnt_com.aftrtrg<Nsmpl_a))
                        and Cnt_com.vld<vld_MAX;
          if (trg='1' and en_rtrg='1' and en_adv='1') then
            BuildNextState_com(TRG_ASSRTD, NextState_com, Cnt_com, NxtCnt_com, vld);                 
          elsif CONDITION then
            BuildNextState_com(CNT_AFTR_TRG, NextState_com, Cnt_com, NxtCnt_com, vld);                 
          elsif en_dtime='1' and not(CONDITION) then
            BuildNextState_com(WAIT_DTIME, NextState_com, Cnt_com, NxtCnt_com, vld);                 
          elsif en_dtime='0' and not(CONDITION) then
            BuildNextState_com(WAIT_TRG, NextState_com, Cnt_com, NxtCnt_com, vld);                 
          else
            REPORT LF & "FAULT CONDITION in process 'p_SM_com_comb' with State_com='" & t_State_com'IMAGE(State_com) &",....)" & LF SEVERITY FAILURE;
            BuildNextState_com(WAIT_TRG, NextState_com, Cnt_com, NxtCnt_com, vld);                 
          end if;
        when WAIT_DTIME =>
          CONDITION := Cnt_com.dtime<DeadTime;  
          if CONDITION then
            BuildNextState_com(WAIT_DTIME, NextState_com, Cnt_com, NxtCnt_com, vld);                 
          elsif not CONDITION then
            BuildNextState_com(WAIT_TRG, NextState_com, Cnt_com, NxtCnt_com, vld);                 
          else
            REPORT LF & "FAULT CONDITION in process 'p_SM_com_comb' with State_com='" & t_State_com'IMAGE(State_com) &",....)" & LF SEVERITY FAILURE;
            BuildNextState_com(WAIT_TRG, NextState_com, Cnt_com, NxtCnt_com, vld);                 
          end if;               
         when OTHERS =>
           REPORT LF & "FAULT CONDITION in process 'p_SM_com_comb' with State_com='" & t_State_com'IMAGE(State_com) &",....)" & LF SEVERITY FAILURE;
           BuildNextState_com(WAIT_TRG, NextState_com, Cnt_com, NxtCnt_com, vld);                 
      end case;    
    -- out
    valid_o <= vld; 
  end process;
  -----------------------------------------

--    all_valids_o(15 downto 0) <= thrsld_ch(0);
  all_valids_o <= mask_trg_ch(trgs_ch'RANGE) AND trgs_ch;

end architecture;
