library IEEE;
use IEEE.STD_LOGIC_1164.ALL;
use IEEE.NUMERIC_STD.ALL;
use IEEE.FIXED_PKG.ALL;
use work.emp_data_types.all;

entity TrackReconstructor is
    port (
        clk_p : in std_logic;
        reset : in std_logic;
        stubs_in : in lword
    );
end TrackReconstructor;

architecture Behavioral of TrackReconstructor is

    signal cache_pointer : integer := 0;
    
    type tStub is record
        valid   : std_logic;
        link_id : integer;
        cic_id  : integer;
        bx      : unsigned(11 downto 0);
        cbc_id  : std_logic_vector(2 downto 0);
        address : std_logic_vector(7 downto 0);
        strip   : signed(10 downto 0);
        y       : sfixed(10 downto -3);
        x       : sfixed(1 downto -1);
        bend    : std_logic_vector(3 downto 0);
    end record tStub;
    constant cNullStub : tStub := ('0', 0, 0, (others => '0'), (others => '0'), (others => '0'), (others => '0'), (others => '0'), (others => '0'), (others => '0'));

    type tStubDelta is record
        valid : std_logic;
        bx : unsigned(11 downto 0);
        dy : sfixed(11 downto -3);
        dx : sfixed(2 downto -1);
        dz : integer;
    end record;
    constant cNullStubDelta : tStubDelta := ('0', (others => '0'), (others => '0'), (others => '0'), 0);

    type tCBCLinkMap is array(0 to 7) of std_logic_vector(2 downto 0);
    constant cCBCLinkMap : tCBCLinkMap := ("111", "110", "101", "100", "000", "001", "010", "011"); -- (-1, -2, -3, -4, 0, 1, 2, 3)
    
    type tUnboundedStubArray is array(integer range <>) of tStub;
    subtype tStubCache is tUnboundedStubArray(0 to 7);
    subtype tStubPair is tUnboundedStubArray(0 to 1);
    type tStubPairArray is array(0 to 7) of tStubPair;

    signal stub_cache : tStubCache := (others => cNullStub);
    signal stub_pair_array : tStubPairArray := (others => (others => cNullStub));  

    signal processing_counter                     : integer   := 0;
    signal processing_start                       : std_logic := '0';
    signal pair_gen_pointer_1, pair_gen_pointer_2 : integer   := 0;
    signal pair_generation_complete               : std_logic := '0';
    signal stub_pair_pointer                      : integer   := 0;

    signal stub_delta : tStubDelta := cNullStubDelta;

begin
    --==============================--
    pFormatting : process(clk_p)
    --==============================--
        variable strip_v : std_logic_vector(10 downto 0) := (others => '0');
        variable formatted_stub : tStub := cNullStub;
    begin
        if rising_edge(clk_p) then
            if reset = '1' then
                stub_cache <= (others => cNullStub);
                cache_pointer <= 0;
            else
                if stubs_in.valid = '1' then                   
                    formatted_stub.valid   := stubs_in.valid;
                    formatted_stub.link_id := to_integer(unsigned(stubs_in.data(63 downto 60)));
                    formatted_stub.cic_id  := to_integer(unsigned(stubs_in.data(46 downto 46)));
                    formatted_stub.bx      := unsigned(stubs_in.data(35 downto 24)) + unsigned(stubs_in.data(17 downto 15));
                    formatted_stub.cbc_id  := cCBCLinkMap(to_integer(unsigned(stubs_in.data(14 downto 12))));
                    formatted_stub.address := stubs_in.data(11 downto 4);
                    strip_v                := cCBCLinkMap(to_integer(unsigned(stubs_in.data(14 downto 12)))) & stubs_in.data(11 downto 4);
                    formatted_stub.strip   := signed(strip_v);
                    formatted_stub.y       := to_sfixed(strip_v & "100", 10, -3);
                    if formatted_stub.cic_id = 0 then
                        formatted_stub.x := to_sfixed(-0.5, 1, -1);
                    elsif formatted_stub.cic_id = 1 then
                        formatted_stub.x := to_sfixed(0.5, 1, -1);
                    end if;
                    formatted_stub.bend := stubs_in.data(3 downto 0);
                    
                    stub_cache(cache_pointer) <= formatted_stub;
                    cache_pointer <= cache_pointer + 1;
                else
                    formatted_stub := cNullStub;
                end if;
            end if;
        end if;
    end process pFormatting;
    
    --==============================--
    pProcessingCounter : process(clk_p)
    --==============================--
    begin
        if rising_edge(clk_p) then
            if reset = '1' then
                processing_counter <= 0;
                processing_start   <= '0';
                
            else
                processing_counter <= processing_counter + 1;
                if processing_counter = 8 then
                    processing_start <= '1';
                end if;
            end if;
        end if;
    end process pProcessingCounter;
    
    --==============================--
    pGenStubPairs : process(clk_p)
    --==============================--
        variable num_stubs : integer := 0;
    begin
        if rising_edge(clk_p) then
            if reset = '1' then
                pair_gen_pointer_1       <= 0;
                pair_gen_pointer_2       <= 0;
                pair_generation_complete <= '0';
                stub_pair_array <= (others => (others => cNullStub));
                stub_pair_pointer <= 0;
            else
                num_stubs := cache_pointer;
                if processing_start = '1' and num_stubs > 0 and pair_generation_complete = '0' then
                    if stub_cache(pair_gen_pointer_2).link_id > stub_cache(pair_gen_pointer_1).link_id then
                        stub_pair_array(stub_pair_pointer) <= (stub_cache(pair_gen_pointer_1), stub_cache(pair_gen_pointer_2));
                        stub_pair_pointer <= stub_pair_pointer + 1;
                    end if;
                    if pair_gen_pointer_2 = num_stubs - 1 then
                        pair_gen_pointer_2 <= 0;
                        if pair_gen_pointer_1 = num_stubs - 1 then
                            pair_gen_pointer_1 <= 0;
                            pair_generation_complete <= '1';
                        else
                            pair_gen_pointer_1 <= pair_gen_pointer_1 + 1;
                        end if;
                    else
                        pair_gen_pointer_2 <= pair_gen_pointer_2 + 1;
                    end if;
                end if;
            end if;
        end if;
    end process pGenStubPairs;

    --==============================--
    pCalculateDelta : process(clk_p)
    --==============================--
    begin
        if rising_edge(clk_p) then
            if reset = '1' then
                stub_delta <= cNullStubDelta;
            else
                if pair_generation_complete = '1' then
                    stub_delta.valid <= '1';
                    stub_delta.bx <= stub_pair_array(0)(0).bx;
                    stub_delta.dx <= stub_pair_array(0)(1).x - stub_pair_array(0)(0).x;
                    stub_delta.dy <= stub_pair_array(0)(1).y - stub_pair_array(0)(0).y;
                    stub_delta.dz <= stub_pair_array(0)(1).link_id - stub_pair_array(0)(0).link_id;
                end if;
            end if;
        end if;
    end process;
end Behavioral;
